package im.creep.luabucket2;

import im.creep.luabucket2.listeners.*;
import org.bukkit.Server;
import org.bukkit.event.Event;
import org.bukkit.event.player.PlayerCommandPreprocessEvent;
import org.bukkit.plugin.Plugin;
import org.bukkit.plugin.PluginManager;
import org.luaj.vm2.LuaClosure;
import org.luaj.vm2.LuaError;
import org.luaj.vm2.LuaFunction;
import org.luaj.vm2.LuaValue;
import org.luaj.vm2.lib.jse.CoerceJavaToLua;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * some day I will put some licence text here, but not now. //apiocera
 */
public class EventDispatcher {
    private static EventDispatcher ourInstance = new EventDispatcher();

    private List<Event.Type> registered_events = new ArrayList<Event.Type>();
    private Map<Event.Type, ArrayList<LuaFunction>> event_callbacks = new HashMap<Event.Type, ArrayList<LuaFunction>>();

    private Map<String, LuaFunction> command_callbacks = new HashMap<String, LuaFunction>();

    private ArrayList<Integer> scheduled_tasks = new ArrayList<Integer>();


    private Plugin thisPlugin = null;
    private PluginManager pluginManager;

    private RunnerBlockListener blockListener;
    private RunnerEntityListener entityListener;
    private RunnerInventoryListener inventoryListener;
    private RunnerPlayerListener playerListener;
    private RunnerServerListener serverListener;
    private RunnerVehicleListener vehicleListener;
    private RunnerWeatherListener weatherListener;
    private RunnerWorldListener worldListener;

    public static EventDispatcher getInstance() {
        return ourInstance;
    }

    private EventDispatcher() {
        blockListener = new RunnerBlockListener(this);
        entityListener = new RunnerEntityListener(this);
        inventoryListener = new RunnerInventoryListener(this);
        playerListener = new RunnerPlayerListener(this);
        serverListener = new RunnerServerListener(this);
        vehicleListener = new RunnerVehicleListener(this);
        weatherListener = new RunnerWeatherListener(this);
        worldListener = new RunnerWorldListener(this);
    }


    // *************** Lua API ********************
    public boolean registerListener(Event.Type ev, LuaFunction callback) {
        if (!(callback instanceof LuaFunction) && !(callback instanceof LuaClosure)) {
            System.out.println("Not a LuaFunction or a LuaClosure.");
            return false;
        }

        if (!event_callbacks.containsKey(ev)) {
            ArrayList<LuaFunction> new_callbacks = new ArrayList<LuaFunction>();
            new_callbacks.add(callback);
            event_callbacks.put(ev, new_callbacks);

            if (!registered_events.contains(ev)) {
                registerInManager(ev, Event.Priority.Normal);
            }
        } else {
            event_callbacks.get(ev).add(callback);
        }
        return true;
    }

    public boolean registerCommand(String command, LuaFunction callback) {
        if (!(callback instanceof LuaFunction) && !(callback instanceof LuaClosure)) {
            System.out.println("Not a LuaFunction or a LuaClosure.");
            return false;
        }

        if (command_callbacks.containsKey(command)) {
            return false;
        }

        command_callbacks.put(command, callback);
        return true;
    }

    public Server getServer(){
        return thisPlugin.getServer();
    }


    public int scheduleSyncDelayedTask(final LuaFunction lf, int delay){
        Integer tId = thisPlugin.getServer().getScheduler().scheduleSyncDelayedTask(thisPlugin, new Runnable() {
            public void run() {
                try {
                    lf.call();
                } catch (LuaError le){
                    Logger.getLogger("minecraft").log(Level.WARNING, "[LuaBucket]: " + le.getMessage());
                }
            }
        }, delay);

        scheduled_tasks.add(tId);
        return tId;
    }

    public int scheduleAsyncDelayedTask(final LuaFunction lf, int delay){
        Integer tId = thisPlugin.getServer().getScheduler().scheduleAsyncDelayedTask(thisPlugin, new Runnable() {
            public void run() {
                try {
                    lf.call();
                } catch (LuaError le){
                    Logger.getLogger("minecraft").log(Level.WARNING, "[LuaBucket]: " + le.getMessage());
                }
            }
        }, delay);

        scheduled_tasks.add(tId);
        return tId;
    }

    public void cancelTask(int tId){
        thisPlugin.getServer().getScheduler().cancelTask(tId);
    }



    // **************** Callbacks for Bukkit ****************
    public void raiseEvent(Event ev) {
        if (event_callbacks.get(ev.getType()) == null) return;

        for (LuaFunction fun : event_callbacks.get(ev.getType())) {
            try {
                fun.call(CoerceJavaToLua.coerce(ev));
            } catch (LuaError le) {
                log(Level.SEVERE, le.getMessage());
            }
        }
    }

    public boolean parseCommand(PlayerCommandPreprocessEvent ev) {
        for (Map.Entry<String, LuaFunction> commandListener : command_callbacks.entrySet()) {
            if (ev.getMessage().equals("/" + commandListener.getKey()) || ev.getMessage().startsWith("/" + commandListener.getKey() + " ")) {
                try {
                    commandListener.getValue().call(CoerceJavaToLua.coerce(ev.getPlayer()), CoerceJavaToLua.coerce(ev.getMessage()));
                } catch (LuaError le) {
                    log(Level.SEVERE, le.getMessage());
                }
                return true;
            }
        }

        return false;
    }

    // **************** Not-so-interesting code *************
    public void reset(){
        for (Integer tId: scheduled_tasks){
            thisPlugin.getServer().getScheduler().cancelTask(tId);
        }

        scheduled_tasks.clear();
        event_callbacks.clear();
        command_callbacks.clear();
    }

    public void setPlugin(Plugin thisPlugin) {
        this.thisPlugin = thisPlugin;
        this.pluginManager = thisPlugin.getServer().getPluginManager();
    }

    private void registerInManager(Event.Type Type, Event.Priority Priority) {
        switch (Type.getCategory()) {
            case BLOCK:
                pluginManager.registerEvent(Type, blockListener, Priority, thisPlugin);
                break;
            case ENTITY:
                pluginManager.registerEvent(Type, entityListener, Priority, thisPlugin);
                break;
            case INVENTORY:
                pluginManager.registerEvent(Type, inventoryListener, Priority, thisPlugin);
                break;
            case PLAYER:
                pluginManager.registerEvent(Type, playerListener, Priority, thisPlugin);
                break;
            case SERVER:
                pluginManager.registerEvent(Type, serverListener, Priority, thisPlugin);
                break;
            case VEHICLE:
                pluginManager.registerEvent(Type, vehicleListener, Priority, thisPlugin);
                break;
            case WEATHER:
                pluginManager.registerEvent(Type, weatherListener, Priority, thisPlugin);
                break;
            case LIVING_ENTITY:
                pluginManager.registerEvent(Type, entityListener, Priority, thisPlugin);
                break;
            case WORLD:
                pluginManager.registerEvent(Type, worldListener, Priority, thisPlugin);
                break;
            case MISCELLANEOUS:
                break;
        }
    }

    private void log(Level level, String message) {
        Logger lg = Logger.getLogger("minecraft");
        lg.log(level, "[LuaBucket]: " + message);
    }
}
